#ifndef _COLLISION_DETECTION_H
#define _COLLISION_DETECTION_H

#include <ostream>
#include <list>

#include "mathvector.h"
#include "quaternion.h"
#include "vertexarray.h"
#include "aabb.h"
#include "aabb_space_partitioning.h"

#include "btBulletCollisionCommon.h"

struct	MultipleRayResultCallback : public btCollisionWorld::RayResultCallback
{
	MultipleRayResultCallback(const btVector3& rayFromWorld,const btVector3& rayToWorld) : m_rayFromWorld(rayFromWorld), m_rayToWorld(rayToWorld) {}
	btVector3	m_rayFromWorld;//used to calculate hitPointWorld from hitFraction
	btVector3	m_rayToWorld;
	
	std::list <btCollisionWorld::LocalRayResult> results;
	
	virtual	btScalar addSingleResult(btCollisionWorld::LocalRayResult& rayResult,bool normalInWorldSpace)
	{
		m_closestHitFraction = rayResult.m_hitFraction;
		results.push_back(rayResult);
		
		if (!normalInWorldSpace)
		{
			///need to transform normal into worldspace
			results.back().m_hitNormalLocal = rayResult.m_collisionObject->getWorldTransform().getBasis()*rayResult.m_hitNormalLocal;
		}
		
		return 1.0;
	}
};

struct COLLISION_OBJECT_VERTEX
{
	float vertex[3];
};

struct COLLISION_OBJECT_FACE
{
	int index[3];
};

class COLLISION_OBJECT_SETTINGS
{
	private:
		bool mapobject;
		bool playerobject;
		bool roadobject;
		void * objid;
	
	public:
		COLLISION_OBJECT_SETTINGS() : mapobject(false),playerobject(false),roadobject(false),objid(NULL) {}
		void SetMapObject(const bool newmapobject) {mapobject = newmapobject;}
		void SetPlayerObject(const bool newplayerobject) {playerobject = newplayerobject;}
		void SetRoadObject(const bool newroadobject) {roadobject = newroadobject;}
		void SetObjectID(void * newobjid) {objid = newobjid;}
	
		bool IsMapObject() const {return mapobject;}
		bool IsPlayerObject() const {return playerobject;}
		bool IsRoadObject() const {return roadobject;}
		const void * ObjID() const {return objid;}
};

class COLLISION_OBJECT
{
	private:
		btCollisionObject id;
		btBvhTriangleMeshShape * trimesh;
		btTriangleIndexVertexArray * trimesh_varray;
	
		bool loaded;
		
		AABB <float> bbox;
	
		COLLISION_OBJECT_SETTINGS settings;

	public:
		COLLISION_OBJECT() : trimesh(NULL),trimesh_varray(NULL),loaded(false) {}
		~COLLISION_OBJECT() {DeInit();}
	
		AABB <float> & GetBBOX() {return bbox;}
		btCollisionObject & GetBulletObject() {return id;}
	
		void InitTrimesh(const float * vertices, int vstride, int vcount, const int * faces, int fcount, int istride, const float * normals, const COLLISION_OBJECT_SETTINGS & objsettings);
		void InitTrimesh(const VERTEXARRAY & varray, const COLLISION_OBJECT_SETTINGS & objsettings);
		void DeInit();
		//dGeomID GetID() const {return id;}
		void SetPosition(const MATHVECTOR <float, 3> & newpos);
		void SetQuaternion(const QUATERNION <float> & newquat);
	
		void SetMapObject(const bool newmapobject) {settings.SetMapObject(newmapobject);}
		void SetPlayerObject(const bool newplayerobject) {settings.SetPlayerObject(newplayerobject);}
		void SetObjectID(void * newobjid) {settings.SetObjectID(newobjid);}
		
		bool IsMapObject() const {return settings.IsMapObject();}
		bool IsPlayerObject() const {return settings.IsPlayerObject();}
		bool IsRoadObject() const {return settings.IsRoadObject();}
		const void * ObjID() const {return settings.ObjID();}
};

class COLLISION_CONTACT
{
	private:
		MATHVECTOR <float, 3> position;
		MATHVECTOR <float, 3> normal;
		float depth;
		COLLISION_OBJECT * col1, * col2;

	public:
		COLLISION_CONTACT() : depth(0), col1(NULL), col2(NULL) {}
		//void SetFromODEContactGeom(dContactGeom odecontactgeom, const PHYSICSCOLLISIONSETTINGS & colsettings);
		//void SetFromBulletContact(btCollisionWorld::LocalRayResult * contact, const PHYSICSCOLLISIONSETTINGS & colsettings);
		const MATHVECTOR <float, 3> & GetContactPosition() const {return position;}
		const MATHVECTOR <float, 3> & GetContactNormal() const {return normal;}
		float GetContactDepth() const {return depth;}
		bool operator<(const COLLISION_CONTACT& other) {return depth < other.depth;}
		COLLISION_OBJECT * GetCollidingObject1() {return col1;}
		COLLISION_OBJECT * GetCollidingObject2() {return col2;}
		void Set(const MATHVECTOR <float, 3> & pos, const MATHVECTOR <float, 3> & norm, float d, COLLISION_OBJECT * c1, COLLISION_OBJECT * c2)
		{
			position = pos;
			normal = norm;
			depth = d;
			col1 = c1;
			col2 = c2;
		}
		bool CollideRay(const MATHVECTOR <float, 3> & origin, const MATHVECTOR <float, 3> & direction, const float length, COLLISION_CONTACT & output_contact) const;
};

class COLLISION_SETTINGS
{
	private:
		bool mapcollide;
		bool playercollide;
		bool roadcollide;
		std::list <void *> exception_objectids;
	
	public:
		COLLISION_SETTINGS() : mapcollide(false),playercollide(false),roadcollide(false) {}
		void SetMapCollide(bool newmc) {mapcollide = newmc;}
		void SetPlayerCollide(bool newpc) {playercollide = newpc;}
		void SetRoadCollide(bool newrc) {roadcollide = newrc;}
		void SetExceptionObjectID(void * except_id) {exception_objectids.push_back(except_id);}
		bool GetMapCollide() const {return mapcollide;}
		bool GetPlayerCollide() const {return playercollide;}
		bool GetRoadCollide() const {return roadcollide;}
		const std::list <void *> & GetExceptionObjectIDs() const {return exception_objectids;}
};

class COLLISION_WORLD
{
	private:
		btCollisionWorld * id;
		btDefaultCollisionConfiguration * collisionconfig;
		btCollisionDispatcher * collisiondispatcher;
		//bt32BitAxisSweep3 * collisionbroadphase;
		btSimpleBroadphase * collisionbroadphase;
		
		AABB_SPACE_PARTITIONING_NODE <COLLISION_OBJECT *> colspeedup;
	
		//void GenerateCollisionList(const COLLISION_SETTINGS & settings, std::list <const COLLISION_OBJECT *> & collisionlist) const;
		//void HandleCollision(const PHYSICSCOLLISIONSETTINGS & settings, MultipleRayResultCallback * contact, list <PHYSICSCONTACT> & outputcontactlist) const;
		bool PassesFilter(const COLLISION_SETTINGS & settings, void * checkme) const;

	public:
		COLLISION_WORLD() : id(NULL),collisionconfig(NULL),collisiondispatcher(NULL),collisionbroadphase(NULL) {}
		~COLLISION_WORLD() {DeInit();}
	
		void Init();
		void DeInit();
		
		void DebugPrint(std::ostream & out)
		{
			std::stringstream junk;
			int numobj(0);
			colspeedup.DebugPrint(0, numobj, true, junk);
			assert(id);
			out << "Collision objects: " << numobj << " (internal), " << id->getNumCollisionObjects() << " (external)" << std::endl;
		}
		
		void CollideRay(const MATHVECTOR <float, 3> & position, const MATHVECTOR <float, 3> & direction, const float length, std::list <COLLISION_CONTACT> & outputcontactlist, const COLLISION_SETTINGS & settings) const;
		void CollideBox(const MATHVECTOR <float, 3> & position, const QUATERNION <float> & orientation, const MATHVECTOR <float, 3> & dimensions, std::list <COLLISION_CONTACT> & outputcontactlist, const COLLISION_SETTINGS & settings) const;
		
		///slow delete
		void DeletePhysicsObject(COLLISION_OBJECT * objtodel)
		{
			assert(id);
			assert(objtodel);
			//id->removeCollisionObject(&objtodel->GetBulletObject());
			colspeedup.Delete(objtodel);
		}
		
		///fast delete
		void DeletePhysicsObject(COLLISION_OBJECT * objtodel, AABB <float> & bbox)
		{
			assert(id);
			assert(objtodel);
			//id->removeCollisionObject(&objtodel->GetBulletObject());
			colspeedup.Delete(objtodel, bbox);
		}
		
		void Clear()
		{
			assert(id);
			std::list <COLLISION_OBJECT * *> allobjects;
			colspeedup.GetContainedObjects(allobjects);
			for (std::list <COLLISION_OBJECT **>::iterator i = allobjects.begin(); i != allobjects.end(); i++)
			{
				btCollisionObject * colobj = &(**i)->GetBulletObject();
				assert(colobj);
				//id->removeCollisionObject(colobj);
			}
			colspeedup.Clear();
		}
	
		void OptimizeObjects() {colspeedup.Optimize();}
		
		void AddPhysicsObject(COLLISION_OBJECT & object)
		{
			assert(id);
			COLLISION_OBJECT * obj = &object;
			AABB <float> bbox = object.GetBBOX();
			colspeedup.Add(obj, bbox);
			
			short int group = 1;
			short int mask = 1;
	
			if (object.IsMapObject())
			{
				group = 2;
				mask = 1;
			}
			else if (object.IsPlayerObject())
			{
				group = 4;
				mask = 1;
			}
			else if (object.IsRoadObject())
			{
				group = 8;
				mask = 1;
			}
			
			btCollisionObject * colobj = &object.GetBulletObject();
			assert(colobj);
			//id->addCollisionObject(colobj, group, mask);
		}
};

#endif
