#include "vertexarray.h"
#include "unittest.h"
#include "mathvector.h"

#include <cassert>

QT_TEST(vertexarray_test)
{
	VERTEXARRAY testarray;
	
	const float * ptr;
	int ptrnum;
	float somevec[3];
	somevec[0] = somevec[2] = 0;
	somevec[1] = 1000.0;
	testarray.SetNormals(somevec, 3);
	testarray.GetNormals(ptr, ptrnum);
	QT_CHECK_EQUAL(ptr[1], 1000.0);
	QT_CHECK_EQUAL(ptrnum, 3);
	testarray.SetNormals(NULL, 0);
	testarray.GetNormals(ptr, ptrnum);
	QT_CHECK_EQUAL(ptrnum, 0);
	QT_CHECK_EQUAL(ptr, NULL);
		
	//by similarity, the vertex and face assignment functions are OK if the above normal test is OK
	
	testarray.SetTexCoordSets(2);
	QT_CHECK_EQUAL(testarray.GetTexCoordSets(), 2);
	testarray.GetTexCoords(0, ptr, ptrnum);
	QT_CHECK_EQUAL(ptrnum, 0);
	QT_CHECK_EQUAL(ptr, NULL);
	testarray.SetTexCoords(1, somevec, 2);
	testarray.GetTexCoords(1, ptr, ptrnum);
	QT_CHECK_EQUAL(ptrnum, 2);
	QT_CHECK_EQUAL(ptr[1], 1000.0);
	
	testarray.Clear();
	testarray.SetNormals(somevec, 3);
	VERTEXARRAY otherarray(testarray);
	otherarray.GetNormals(ptr, ptrnum);
	QT_CHECK_EQUAL(ptr[1], 1000.0);
	QT_CHECK_EQUAL(ptrnum, 3);
	otherarray.Clear();
	otherarray.GetNormals(ptr, ptrnum);
	QT_CHECK_EQUAL(ptrnum, 0);
	otherarray = testarray;
	otherarray.GetNormals(ptr, ptrnum);
	QT_CHECK_EQUAL(ptr[1], 1000.0);
	QT_CHECK_EQUAL(ptrnum, 3);
	
	VERTEXARRAY addarray;
	otherarray.Clear();
	testarray.Clear();
	testarray.SetNormals(somevec, 3);
	otherarray.SetNormals(somevec, 3);
	addarray = testarray + otherarray;
	addarray.GetNormals(ptr, ptrnum);
	QT_CHECK_EQUAL(ptrnum, 6);
	QT_CHECK_EQUAL(ptr[1], 1000.0);
	QT_CHECK_EQUAL(ptr[4], 1000.0);
	
	testarray.Clear();
	VERTEXARRAY facearray1, facearray2;
	int someint[3];
	someint[0] = 0;
	someint[1] = 1;
	someint[2] = 2;
	facearray1.SetFaces(someint, 3);
	facearray2.SetFaces(someint, 3);
	testarray = facearray1 + facearray2;
	const int * ptri;
	testarray.GetFaces(ptri, ptrnum);
	QT_CHECK_EQUAL(ptrnum, 6);
	QT_CHECK_EQUAL(ptri[1], 1);
	QT_CHECK_EQUAL(ptri[4], 1);
	testarray.Clear();
	facearray1.SetVertices(somevec, 3);
	facearray2.SetVertices(somevec, 3);
	testarray = facearray1 + facearray2;
	testarray.GetFaces(ptri, ptrnum);
	QT_CHECK_EQUAL(ptrnum, 6);
	QT_CHECK_EQUAL(ptri[1], 1);
	QT_CHECK_EQUAL(ptri[4], 2);
}

void VERTEXARRAY::SetNormals(float * newarray, size_t newarraycount)
{
	//Tried to assign values that aren't in sets of 3
	assert(newarraycount % 3 == 0);
	
	if (newarraycount != normals.size())
	{
		normals.resize(newarraycount);
	}
	
	float * myarray = &(normals[0]);
	for (size_t i = 0; i < newarraycount; ++i)
	{
		myarray[i] = newarray[i];
	}
	
	/*for (int i = 0; i < newarraycount; i++)
	{
		normals[i] = newarray[i];
	}*/
}

void VERTEXARRAY::SetVertices(float * newarray, size_t newarraycount)
{
	//Tried to assign values that aren't in sets of 3
	assert(newarraycount % 3 == 0);
	
	if (newarraycount != vertices.size())
	{
		//cout << "New size: " << newarraycount << " (was " << vertices.size() << ")" << endl;
		vertices.resize(newarraycount);
	}
	
	float * myarray = &(vertices[0]);
	for (size_t i = 0; i < newarraycount; ++i)
	{
		myarray[i] = newarray[i];
	}
	
	/*int i = 0;
	ITERVEC(float, vertices, it)
	{
		*it = newarray[i];
		++i;
	}*/
}

void VERTEXARRAY::SetFaces(int * newarray, size_t newarraycount)
{
	//Tried to assign values that aren't in sets of 3
	assert (newarraycount % 3 == 0);
	
	if (newarraycount != faces.size())
		faces.resize(newarraycount);
	
	
	int * myarray = &(faces[0]);
	for (size_t i = 0; i < newarraycount; ++i)
	{
		myarray[i] = newarray[i];
	}
	
	/*for (int i = 0; i < newarraycount; i++)
	{
		faces[i] = newarray[i];
	}*/
}

void VERTEXARRAY::SetTexCoordSets(int newtcsets)
{
	texcoords.clear();
	texcoords.resize(newtcsets);
}

//set is zero indexed
void VERTEXARRAY::SetTexCoords(size_t set, float * newarray, size_t newarraycount)
{
	//Tried to assign a tex coord set beyond the allocated number of sets
	assert(set < texcoords.size());
	
	//Tried to assign values that aren't in sets of 2
	assert(newarraycount % 2 == 0);
	
	if (texcoords[set].size() != newarraycount)
		texcoords[set].resize(newarraycount);
	
	float * myarray = &(texcoords[set][0]);
	for (size_t i = 0; i < newarraycount; ++i)
	{
		myarray[i] = newarray[i];
	}
	
	/*for (int i = 0; i < newarraycount; i++)
	{
		texcoords[set][i] = newarray[i];
	}*/
}

void VERTEXARRAY::GetNormals(const float * & output_array_pointer, int & output_array_num) const
{
	output_array_num = normals.size();
	output_array_pointer = normals.empty() ? NULL : &normals[0];
}

void VERTEXARRAY::GetVertices(const float * & output_array_pointer, int & output_array_num) const
{
	output_array_num = vertices.size();
	output_array_pointer = vertices.empty() ? NULL : &vertices[0];
}

void VERTEXARRAY::GetFaces(const int * & output_array_pointer, int & output_array_num) const
{
	output_array_num = faces.size();
	//output_array_pointer = faces.empty() ? NULL : &faces[0];
	if (faces.size() < 1)
		output_array_pointer = NULL;
	else
		output_array_pointer = &faces[0];
}

void VERTEXARRAY::GetTexCoords(size_t set, const float * & output_array_pointer, int & output_array_num) const
{
	//Tried to get a tex coord set beyond the allocated number of sets
	assert(set < texcoords.size());
	
	output_array_num = texcoords[set].size();
	output_array_pointer = texcoords[set].empty() ? NULL : &texcoords[set][0];
}

#define COMBINEVECTORS(vname) {out.vname.reserve(vname.size() + v.vname.size());out.vname.insert(out.vname.end(), vname.begin(), vname.end());out.vname.insert(out.vname.end(), v.vname.begin(), v.vname.end());}

VERTEXARRAY VERTEXARRAY::operator+ (const VERTEXARRAY & v) const
{
	VERTEXARRAY out;
	
	//out.normals.reserve(normals.size() + v.normals.size());
	//out.normals.insert(out.normals.end(), normals.begin(), normals.end());
	//out.normals.insert(out.normals.end(), v.normals.begin(), v.normals.end());
	
	int idxoffset = vertices.size()/3;
	
	COMBINEVECTORS(normals)
	COMBINEVECTORS(vertices)
	
	//COMBINEVECTORS(faces)
	out.faces.reserve(faces.size() + v.faces.size());
	out.faces.insert(out.faces.end(), faces.begin(), faces.end());
	//out.faces.insert(out.faces.end(), v.faces.begin(), v.faces.end());
	//size_t startat = out.faces.size();
	for (size_t i = 0; i < v.faces.size(); i++)
	{
		//out.faces[faces.size() + i] = v.faces[i] + idxoffset;
		//out.faces[startat + i] = v.faces[i];
		out.faces.push_back(v.faces[i]+idxoffset);
	}
	
	int maxtcsets = GetTexCoordSets();
	if (v.GetTexCoordSets() > maxtcsets)
		maxtcsets = v.GetTexCoordSets();
	int tcsets1 = GetTexCoordSets();
	int tcsets2 = v.GetTexCoordSets();
	out.SetTexCoordSets(maxtcsets);
	//ITERVEC(vector <float>,texcoords,i)
	for (int i = 0; i < maxtcsets; i++)
	{
		if (i >= tcsets1 && i < tcsets2)
		{
			out.texcoords[i] = v.texcoords[i];
		}
		else if (i < tcsets1 && i >= tcsets2)
		{
			out.texcoords[i] = texcoords[i];
		}
		else if (i < tcsets1 && i < tcsets2)
		{
			COMBINEVECTORS(texcoords[i])
		}
	}
	
	return out;
}

void VERTEXARRAY::Add(float * newnorm, int newnormcount, float * newvert, int newvertcount,
			int * newfaces, int newfacecount,
			float * newtc, int newtccount)
{
	int idxoffset = vertices.size()/3;
	
	//add normals
	{
		int newcount = newnormcount;
		int origsize = normals.size();
		
		//Tried to assign values that aren't in sets of 3
		assert(newcount % 3 == 0);
		
		int newsize = origsize + newcount;
		if (newcount > 0)
			normals.resize(newsize);
		float * myarray = &(normals[0]);
		for (int i = 0; i < newcount; ++i)
		{
			myarray[i+origsize] = newnorm[i];
		}
	}
	
	//add verts
	{
		int newcount = newvertcount;
		int origsize = vertices.size();
		
		//Tried to assign values that aren't in sets of 3
		assert(newcount % 3 == 0);
		
		int newsize = origsize + newcount;
		if (newcount > 0)
			vertices.resize(newsize);
		float * myarray = &(vertices[0]);
		for (int i = 0; i < newcount; ++i)
		{
			myarray[i+origsize] = newvert[i];
		}
	}
	
	//add faces
	{
		int newcount = newfacecount;
		int origsize = faces.size();
		
		//Tried to assign values that aren't in sets of 3
		assert (newcount % 3 == 0);
			
		int newsize = origsize + newcount;
		if (newcount > 0)
			faces.resize(newsize);
		int * myarray = &(faces[0]);
		for (int i = 0; i < newcount; ++i)
		{
			myarray[i+origsize] = newfaces[i] + idxoffset;
		}
	}
	
	//add tex coords
	{
		//This version of VERTEXARRAY::Add assumes texcoordsets == 1
		assert(GetTexCoordSets() == 1);
		
		int newcount = newtccount;
		int origsize = texcoords[0].size();
		
		//Tried to assign values that aren't in sets of 2
		assert(newcount % 2 == 0);
		
		int newsize = origsize + newcount;
		if (newcount > 0)
			texcoords[0].resize(newsize);
		float * myarray = &(texcoords[0][0]);
		for (int i = 0; i < newcount; ++i)
		{
			myarray[i+origsize] = newtc[i];
		}
	}
}

void VERTEXARRAY::SetToBillboard(float x1, float y1, float x2, float y2)
{
	/*glNormal3f( 0.0f, 0.0f, 1.0f);
	glTexCoord2f(0.0f, 0.0f); glVertex3f( 0, -scale,  -scale);
	glTexCoord2f(1.0f, 0.0f); glVertex3f( 0, -scale,  scale);
	glTexCoord2f(1.0f, 1.0f); glVertex3f( 0,  scale,  scale);
	glTexCoord2f(0.0f, 1.0f); glVertex3f( 0,  scale,  -scale);*/
	
	int bfaces[6];
	bfaces[0] = 0;
	bfaces[1] = 1;
	bfaces[2] = 2;
	bfaces[3] = 0;
	bfaces[4] = 2;
	bfaces[5] = 3;
	SetFaces(bfaces, 6);
	
	float normals[12];
	for (int i = 0; i < 12; i+=3)
	{
		normals[i] = 0;
		normals[i+1] = 0;
		normals[i+2] = 1;
	}
	SetNormals(normals, 12);
	
	float verts[12];
	
	/*verts[0] = verts[3] = verts[6] = verts[9] = 0.0;
	verts[1] = verts[2] = verts[4] = verts[11] = -scale;
	verts[5] = verts[7] = verts[8] = verts[10] = scale;*/
	
	//build this:
	//x1, y1, 0
	//x2, y1, 0
	//x2, y2, 0
	//x1, y2, 0
	verts[2] = verts[5] = verts[8] = verts[11] = 0.0;
	verts[0] = verts[9] = x1;
	verts[3] = verts[6] = x2;
	verts[1] = verts[4] = y1;
	verts[7] = verts[10] = y2;
	SetVertices(verts, 12);
	
	float tc[8];
	tc[0] = tc[1] = tc[3] = tc[6] = 0.0;
	tc[2] = tc[4] = tc[5] = tc[7] = 1.0;
	SetTexCoordSets(1);
	SetTexCoords(0, tc, 8);
}

void VERTEXARRAY::SetVertexData2DQuad(float x1, float y1, float x2, float y2, float u1, float v1, float u2, float v2, float * vcorners, float * uvs, int * bfaces, int faceoffset) const
{
	vcorners[0] = x1;
	vcorners[1] = y1;
	vcorners[2] = 0;
	vcorners[3] = x2;
	vcorners[4] = y1;
	vcorners[5] = 0;
	vcorners[6] = x2;
	vcorners[7] = y2;
	vcorners[8] = 0;
	vcorners[9] = x1;
	vcorners[10] = y2;
	vcorners[11] = 0;

	uvs[0] = u1;
	uvs[1] = v1;
	uvs[2] = u2;
	uvs[3] = v1;
	uvs[4] = u2;
	uvs[5] = v2;
	uvs[6] = u1;
	uvs[7] = v2;
	
	bfaces[0] = faceoffset+0;
	bfaces[1] = faceoffset+2;
	bfaces[2] = faceoffset+1;
	bfaces[3] = faceoffset+0;
	bfaces[4] = faceoffset+3;
	bfaces[5] = faceoffset+2;
}

void VERTEXARRAY::SetTo2DButton(float x, float y, float w, float h, float sidewidth, bool flip)
{
	float vcorners[12*3];
	float uvs[8*3];
	int bfaces[6*3];
			
	//y1 = 1.0 - y1;
	//y2 = 1.0 - y2;
	
	MATHVECTOR <float, 2> corner1;
	MATHVECTOR <float, 2> corner2;
	MATHVECTOR <float, 2> dim;
	dim.Set(w,h);
	MATHVECTOR <float, 2> center;
	center.Set(x,y);
	corner1 = center - dim*0.5;
	corner2 = center + dim*0.5;
	
	float x1 = corner1[0];
	float y1 = corner1[1];
	float x2 = corner2[0];
	float y2 = corner2[1];
	
	if (flip)
	{
		float y3 = y1;
		y1 = y2;
		y2 = y3;
	}
			
	//left
	SetVertexData2DQuad(x1-sidewidth,y1,x1,y2, 0,0,0.5,1, vcorners, uvs, bfaces);
			
	//center
	SetVertexData2DQuad(x1,y1,x2,y2, 0.5,0,0.5,1, &(vcorners[12]), &(uvs[8]), &(bfaces[6]), 4);
			
	//right
	SetVertexData2DQuad(x2,y1,x2+sidewidth,y2, 0.5,0,1,1, &(vcorners[12*2]), &(uvs[8*2]), &(bfaces[6*2]), 8);
			
	SetFaces(bfaces, 6*3);
	SetVertices(vcorners, 12*3);
	SetTexCoordSets(1);
	SetTexCoords(0, uvs, 8*3);
}

void VERTEXARRAY::SetTo2DBox(float x, float y, float w, float h, float marginwidth, float marginheight)
{
	const unsigned int quads = 9;
	float vcorners[12*quads];
	float uvs[8*quads];
	int bfaces[6*quads];
			
	//y1 = 1.0 - y1;
	//y2 = 1.0 - y2;
	
	MATHVECTOR <float, 2> corner1;
	MATHVECTOR <float, 2> corner2;
	MATHVECTOR <float, 2> dim;
	dim.Set(w,h);
	MATHVECTOR <float, 2> center;
	center.Set(x,y);
	corner1 = center - dim*0.5;
	corner2 = center + dim*0.5;
	MATHVECTOR <float, 2> margin;
	margin.Set(marginwidth, marginheight);
	
	//upper left
	SetVertexData2DQuad((corner1-margin)[0],(corner1-margin)[1],corner1[0],corner1[1],
			    0,0,0.5,0.5, vcorners,uvs,bfaces);
	
	//upper center
	SetVertexData2DQuad(corner1[0],(corner1-margin)[1],corner2[0],corner1[1],
			     0.5,0,0.5,0.5,
			     &(vcorners[12*1]),&(uvs[8*1]),&(bfaces[6*1]),4*1);
	
	//upper right
	SetVertexData2DQuad(corner2[0],(corner1-margin)[1],(corner2+margin)[0],corner1[1],
			    0.5,0,1,0.5,
			    &(vcorners[12*2]),&(uvs[8*2]),&(bfaces[6*2]),4*2);
	
	//center left
	SetVertexData2DQuad((corner1-margin)[0],corner1[1],corner1[0],corner2[1],
			    0,0.5,0.5,0.5,
			    &(vcorners[12*3]),&(uvs[8*3]),&(bfaces[6*3]),4*3);
	
	//center center
	SetVertexData2DQuad(corner1[0],corner1[1],corner2[0],corner2[1],
			    0.5,0.5,0.5,0.5,
			    &(vcorners[12*4]),&(uvs[8*4]),&(bfaces[6*4]),4*4);
	
	//center right
	SetVertexData2DQuad(corner2[0],corner1[1],(corner2+margin)[0],corner2[1],
			    0.5,0.5,1,0.5,
			    &(vcorners[12*5]),&(uvs[8*5]),&(bfaces[6*5]),4*5);
	
	//lower left
	SetVertexData2DQuad((corner1-margin)[0],corner2[1],corner1[0],(corner2+margin)[1],
			    0,0.5,0.5,1,
			    &(vcorners[12*6]),&(uvs[8*6]),&(bfaces[6*6]),4*6);
	
	//lower center
	SetVertexData2DQuad(corner1[0],corner2[1],corner2[0],(corner2+margin)[1],
			    0.5,0.5,0.5,1,
			    &(vcorners[12*7]),&(uvs[8*7]),&(bfaces[6*7]),4*7);
	
	//lower right
	SetVertexData2DQuad(corner2[0],corner2[1],(corner2+margin)[0],(corner2+margin)[1],
			    0.5,0.5,1,1,
			    &(vcorners[12*8]),&(uvs[8*8]),&(bfaces[6*8]),4*8);
	
	SetFaces(bfaces, 6*quads);
	SetVertices(vcorners, 12*quads);
	SetTexCoordSets(1);
	SetTexCoords(0, uvs, 8*quads);
}

